<?php

class eaccCache extends Cache {
  /**
   * page_fast_cache
   *   This tells CacheRouter to use page_fast_cache.
   *
   *   @return bool TRUE
   */
  function page_fast_cache() {
    return TRUE;
  }

  /**
   * get()
   *   Return item from cache if it is available.
   *
   * @param string $key
   *   The key to fetch.
   * @return mixed object|bool
   *   Returns either the cache object or FALSE on failure
   */
  function get($key) {
    $cache = parent::get($key);
    if (isset($cache)) {
      return $cache;
    }

    $cache = eaccelerator_get($this->key($key));
    if (!is_null($cache)) {
      $cache = unserialize($cache);
    }
    else {
      $cache = 0;
    }
    parent::set($key, $cache);
    return $cache;
  }

  /**
   * set()
   *   Add item into cache.
   *
   * @param string $key
   *   The key to set.
   * @param string $value
   *   The data to store in the cache.
   * @param string $expire
   *   The time to expire in seconds.
   * @param string $headers
   *   The page headers.
   * @return bool
   *   Returns TRUE on success or FALSE on failure
   */
  function set($key, $value, $expire = CACHE_PERMANENT, $headers = NULL) {
    if (empty($key)) {
      return FALSE;
    }
    $return = FALSE;
    // Create new cache object.
    $cache = new stdClass;
    $cache->cid = $key;
    $cache->created = time();
    $cache->expire = $expire;
    $cache->headers = $headers;

    $cache->serialized = FALSE;
    $cache->data = $value;

    //Store temporary as permanent - clean in flush
    if ($expire == CACHE_TEMPORARY || $expire == CACHE_PERMANENT) {
      $ttl = 0;
    }
    else {
      $ttl = $expire - $cache->created;
    }

    // Get full key for storage and set it to $expire so we can keep track of the bin
    $full_key = $this->key($key);

    // Attempt to store full key and value
    if (eaccelerator_put($full_key, serialize($cache), $ttl)) {
      // Update static cache
      parent::set($key, $cache);
      $return = TRUE;

      // Get lookup table to be able to keep track of bins
      $this->lock();
      $lookup = eaccelerator_get($this->lookup);

      // If the lookup table is empty, initialize table
      if (empty($lookup)) {
        $lookup = array();
      }
      else {
      	$lookup = unserialize($lookup);
      }

      $lookup[$full_key] = $expire;

      // Resave the lookup table (even on failure)
      eaccelerator_put($this->lookup, serialize($lookup));

      // Remove lock.
      $this->unlock();
    }
    return $return;
  }

  /**
   * delete()
   *   Remove item from cache.
   *
   * @param string $key
   *   The key to delete.
   * @return mixed object|bool
   *   Returns either the cache object or FALSE on failure
   */
  function delete($key) {
    if (substr($key, -1, 1) == '*') {
      //Clear static array cache (no wildcards)
      parent::flush();

      $this->lock();
      $lookup = eaccelerator_get($this->lookup);
      if (!is_null($lookup)) {
      	$lookup = unserialize($lookup);
      }
      if (is_array($lookup)) {
        if ($key == '*') {
          //Fast clean of lookup
          eaccelerator_put($this->lookup, serialize(array()));
          $this->unlock();
          foreach ($lookup as $k => $v) {
            eaccelerator_rm($k);
          }
          return TRUE;
        }
        else {
          $key = $this->key(substr($key, 0, strlen($key) - 1));
          foreach ($lookup as $k => $v) {
            if (strpos($k, $key) === 0) {
            //if (substr($k, 0, strlen($key) - 1)) {
              eaccelerator_rm($k);
              unset($lookup[$k]);
            }
          }
        }
      }
      else {
      	$lookup = array();
      }
      eaccelerator_put($this->lookup, serialize($lookup));
      $this->unlock();
    }
    else {
      //Remove only key - clean $lookup on flush
      parent::delete($key);
      return eaccelerator_rm($this->key($key));
    }
  }

  /**
   * flush()
   *   Flush the entire cache.
   *
   * @param none
   * @return mixed bool
   *   Returns TRUE
   */
  function flush() {
    parent::flush();
    $this->lock();

    // Get lookup table to be able to keep track of bins
    $lookup = eaccelerator_get($this->lookup);

    // If the lookup table is empty, remove lock and return
    if (empty($lookup)) {
      $this->unlock();
      return TRUE;
    }
    $lookup = unserialize($lookup);

    // Cycle through keys and remove each entry from the cache
    if (is_array($lookup)) {
      foreach ($lookup as $k => $v) {
        if ($v == CACHE_TEMPORARY) {
          eaccelerator_rm($k);
          unset($lookup[$k]);
        }
        elseif (is_null(eaccelerator_get($k))) {
          unset($lookup[$k]);
        }
      }
    }
    else {
    	$lookup = array();
    }

    // Resave the lookup table (even on failure)
    eaccelerator_put($this->lookup, serialize($lookup));

    // Remove lock
    $this->unlock();
    eaccelerator_gc();
    return TRUE;
  }

  /**
   * lock()
   *   lock the cache from other writes.
   *
   * @param none
   * @return string
   *   Returns TRUE on success, FALSE on failure
   */
  function lock() {
    return eaccelerator_lock($this->lock);
  }

  /**
   * unlock()
   *   lock the cache from other writes.
   *
   * @param none
   * @return bool
   *   Returns TRUE on success, FALSE on failure
   */
  function unlock() {
    return  eaccelerator_unlock($this->lock);
  }
}