<?php
class memcacheCache extends Cache {
  var $settings = array();
  var $memcache;

  /**
   * page_fast_cache
   *   This tells CacheRouter to use page_fast_cache.
   *
   *   @return bool TRUE
   */
  function page_fast_cache() {
    return TRUE;
  }

  function __construct($bin) {
    global $conf;
    // Assign the servers on the following order: bin specific -> default specific -> localhost port 11211
    if (isset($conf['cacherouter'][$bin]['server'])) {
    	$this->settings['servers'] = $conf['cacherouter'][$bin]['server'];
    	$this->settings['compress'] = isset($conf['cacherouter'][$bin]['compress']) ? MEMCACHE_COMPRESSED : 0;
      $this->settings['shared'] = isset($conf['cacherouter'][$bin]['shared']) ?
                                  $conf['cacherouter'][$bin]['shared'] : TRUE;
    }
    else {
      if (isset($conf['cacherouter']['default']['server'])) {
        $this->settings['servers'] = $conf['cacherouter']['default']['server'];
        $this->settings['compress'] = isset($conf['cacherouter']['default']['compress']) ? MEMCACHE_COMPRESSED : 0;
        $this->settings['shared'] = isset($conf['cacherouter']['default']['shared']) ?
                                    $conf['cacherouter']['default']['shared'] : TRUE;
      }
      else {
        $this->settings['servers'] = array('localhost:11211');
        $this->settings['compress'] = 0;
        $this->settings['shared'] = TRUE;
      }
    }

    parent::__construct($bin);

    $this->connect();
  }

  /**
   * get()
   *   Return item from cache if it is available.
   *
   * @param string $key
   *   The key to fetch.
   * @return mixed object|bool
   *   Returns either the cache object or FALSE on failure
   */
  function get($key) {
    // Attempt to pull from static cache.
    $cache = parent::get($key);
    if (isset($cache)) {
      return $cache;
    }

    // Get from memcache
    $cache = $this->memcache->get($this->key($key));
    if ($cache === FALSE) {
      $cache = 0;
    }
    parent::set($key, $cache);
    return $cache;
  }

  /**
   * set()
   *   Add item into cache.
   *
   * @param string $key
   *   The key to set.
   * @param string $value
   *   The data to store in the cache.
   * @param string $expire
   *   The time to expire in seconds.
   * @param string $headers
   *   The page headers.
   * @return bool
   *   Returns TRUE on success or FALSE on failure
   */
  function set($key, $value, $expire = CACHE_PERMANENT, $headers = NULL) {
    if (empty($key)) {
      return FALSE;
    }
    $return = FALSE;
    // Create new cache object.
    $cache = new stdClass;
    $cache->cid = $key;
    $cache->created = time();
    $cache->expire = $expire;
    $cache->headers = $headers;

    $cache->serialized = FALSE;
    $cache->data = $value;

    //Store temporary as permanent - clean in flush
    if ($expire == CACHE_TEMPORARY || $expire == CACHE_PERMANENT) {
      $ttl = 0;
    }
    else {
      $ttl = $expire - $cache->created;
    }

    // Get full key for storage and set it to $expire so we can keep track of the bin
    $full_key = $this->key($key);
    // Attempt to store full key and value
    if ($this->memcache->set($full_key, $cache, $this->settings['compress'], $ttl)) {
      // Update static cache
      parent::set($key, $cache);
      $return = TRUE;

      if ($this->settings['shared']) {
        // Get lookup table to be able to keep track of bins
        $this->lock();
        $lookup = $this->memcache->get($this->lookup);

        // If the lookup table is empty, initialize table
        if (empty($lookup)) {
          $lookup = array();
        }

        $lookup[$full_key] = $expire;

        // Resave the lookup table (even on failure)
        $this->memcache->set($this->lookup, $lookup, $this->settings['compress']);

        // Remove lock.
        $this->unlock();
      }
    }
    return $return;
  }

  /**
   * delete()
   *   Remove item from cache.
   *
   * @param string $key
   *   The key to delete.
   * @return mixed object|bool
   *   Returns either the cache object or FALSE on failure
   */
  function delete($key) {
    if (substr($key, -1, 1) == '*') {

      if ($this->settings['shared']) {
        //Clear static array cache (no wildcards)
        parent::flush();

        // Get lookup table to be able to keep track of bins
        $this->lock();
        $lookup = $this->memcache->get($this->lookup);
        if (is_array($lookup)) {
          if ($key == '*') {
            //Fast clean of lookup
            $this->memcache->set($this->lookup, array(), $this->settings['compress']);
            $this->unlock();
            foreach ($lookup as $k => $v) {
              $this->memcache->delete($k);
            }
            return TRUE;
          }
          else {
            $key = $this->key(substr($key, 0, strlen($key) - 1));
            foreach ($lookup as $k => $v) {
              if (strpos($k, $key) === 0) {
                $this->memcache->delete($k);
                unset($lookup[$k]);
              }
            }
          }
        }
        else {
        	$lookup = array();
        }
        $this->memcache->set($this->lookup, $lookup, $this->settings['compress']);
        $this->unlock();
      }
      else {
      	$this->flush();
      }
    }
    else {
      //Remove only key - clean $lookup on flush
      parent::delete($key);
      return $this->memcache->delete($this->key($key));
    }
  }

  /**
   * flush()
   *   Flush the entire cache.
   *
   * @param none
   * @return mixed bool
   *   Returns TRUE
   */
  function flush() {
    // Flush static cache
    parent::flush();

    // If this is a shared cache, we need to cycle through the lookup table and remove individual
    // items directly
    if ($this->settings['shared']) {
      $this->lock();

      // Get lookup table to be able to keep track of bins
      $lookup = $this->memcache->get($this->lookup);

      // If the lookup table is empty, remove lock and return
      if (empty($lookup)) {
        $this->unlock();
        return TRUE;
      }

      // Cycle through keys and remove each entry from the cache
      if (is_array($lookup)) {
        foreach ($lookup as $k => $v) {
          if ($v == CACHE_TEMPORARY) {
            $this->memcache->delete($k);
            unset($lookup[$k]);
          }
          elseif ($this->memcache->get($k) === FALSE) {
            unset($lookup[$k]);
          }
        }
      }
      else {
      	$lookup = array();
      }

      // Resave the lookup table (even on failure)
      $this->memcache->set($this->lookup, $lookup, $this->settings['compress'], 0);

      // Remove lock
      $this->unlock();
    }
    else {
      // Flush memcache - clean all stored data
      return $this->memcache->flush();
    }
    return TRUE;
  }

  /**
   * lock()
   *   lock the cache from other writes.
   *
   * @param none
   * @return string
   *   Returns TRUE on success, FALSE on failure
   */
  function lock() {
    // Lock once by trying to add lock file, if we can't get the lock, we will loop
    // for 3 seconds attempting to get lock.  If we still can't get it at that point,
    // then we give up and return FALSE.
    if ($this->memcache->add($this->lock, $this->settings['compress'], 0) === FALSE) {
      $time = time();
      while ($this->memcache->add($this->lock, $this->settings['compress'], 0) === FALSE) {
        if (time() - $time >= 3) {
          return FALSE;
        }
      }
    }
    return TRUE;
  }

  /**
   * unlock()
   *   lock the cache from other writes.
   *
   * @param none
   * @return bool
   *   Returns TRUE on success, FALSE on failure
   */
  function unlock() {
    return $this->memcache->delete($this->lock);
  }

  function connect() {
    $this->memcache =& new Memcache;
    foreach ($this->settings['servers'] as $server) {
      list($host, $port) = explode(':', $server);
      if (!$this->memcache->addServer($host, $port)) {
        drupal_set_message("Unable to connect to memcache server $host:$port");
      }
    }
  }

  function close() {
    $this->memcache->close();
  }
}